#!/usr/bin/env python

"""
This script calculates the free energy of the WCA system along the T=0.80
isotherm by thermodynamic integration. It does not fit into the scheme used
by the other tests.
"""

import sys
sys.path.insert(0,"../../../Swig")
sys.path.insert(0,"../../../Python")
sys.path.insert(0,"../../../Tools")

import rumdSimulation, rumd, rumd_tools
# it is essential that there is no doubt about which version is actually run!
print "File used to import the rumd module",rumd.__file__
print "File used to import the rumdSimulation module",rumdSimulation.__file__
print "File used to import the rumd_tools module",rumd_tools.__file__


import os
import numpy
import string

if os.environ.has_key("PBS_O_WORKDIR"):
    workingDir = os.environ["PBS_O_WORKDIR"]
    os.chdir(workingDir)
    sys.path.append(".")
                

from Scientific.Functions.LeastSquares import polynomialLeastSquaresFit, leastSquaresFit
from rumdSimulation import rumd, rumdSimulation
import rumd_tools

class WCA_FreeEnergy:
    def __init__(self):
        self.timeStep = 0.002
        self.rhoMin = 0.05
        self.rhoMax = 0.90
        self.rhoStep = 0.05

        self.n_equil_steps = 25000
        self.n_run_steps = 50000
        self.startConfigFile = "start.xyz.gz"
        self.momentum_reset_interval = 5000
        self.potLJCutoff = 4.0
        self.n_equil_stepsStart = 100000
        self.energyInterval = 50
        
    def CreateSimulation(self):
        sim = rumdSimulation(self.startConfigFile, 32, 8)
        self.sim = sim
        itg = rumd.IntegratorNVT(targetTemperature=self.T, timeStep=self.timeStep)
        sim.SetIntegrator(itg)
        
        self.SetDensity(self.rhoMin)
        

        self.potWCA = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        self.potWCA.SetParams(0,0,1.,1.,pow(2.,1./6))
        sim.SetPotential(self.potWCA)

        self.potLJ = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        self.potLJ.SetParams(0,0,1.,1.,self.potLJCutoff)
        self.vLJ_cut = 4*(1/self.potLJCutoff**12 - 1/self.potLJCutoff**6)

        
        sim.SetOutputScheduling("trajectory", "none" )
        sim.SetOutputScheduling("energies", "linear", interval=self.energyInterval )
        sim.momentum_reset_interval = self.momentum_reset_interval

        sim.SetVerbose(False)

        sim.Run(self.n_equil_stepsStart, suppressAllOutput=True)

        self.alt_pot_calc = rumd.AlternatePotentialCalculator(self.sim.sample, self.potLJ)
        sim.AddExternalCalculator(self.alt_pot_calc)


        
        
    def SetDensity(self, rh):
        # scale to get right density
        nParticles = self.sim.GetNumberOfParticles()
        vol = self.sim.GetVolume()
        currentDensity = nParticles/vol
        scaleFactor = pow(rh/currentDensity, -1./3)
        
        self.sim.ScaleSystem(scaleFactor)



    def MainLoop(self, T):
        self.T = T
        nRuns = int((self.rhoMax - self.rhoMin)/self.rhoStep + 1.e-8) + 1

        self.CreateSimulation()

        outfile = open("outputWCA%.3f.dat" % T, "w")
        outfile.write("# rho, meanW, W0, meanPE\n")

        for rdx in xrange(nRuns):
            
            rho = self.rhoMin + rdx*self.rhoStep
            print "rdx, rho",rdx, rho
            self.SetDensity(rho)

            self.sim.Run(self.n_equil_steps, suppressAllOutput=True)

            self.sim.Run(self.n_run_steps)
 

            rs = rumd_tools.rumd_stats()
            rs.ComputeStats()
            meanVals = rs.GetMeanVals()
            meanPE = meanVals["pe"]
            meanW = meanVals["W"]
            if meanVals.has_key("potLJ_12_6"):
                mean_pe_LJ = meanVals["potLJ_12_6"]

                correction1 = 0.5*(4.*numpy.pi/3*self.potLJCutoff**3*rho - 1)*self.vLJ_cut
                correction2 = -8.*numpy.pi*rho/3./self.potLJCutoff**3
                mean_pe_LJ += correction1 + correction2

                W0 = mean_pe_LJ - meanPE
            else:
                W0 = 0.0
                
            outfile.write("%.3f %.5f %.5f %.5f\n" % (rho, meanW, W0, meanPE))
        outfile.close()

def PolyModel2(params, rh):
    r = 0.0
    for i in range(len(params)):
        r = r + params[i]*numpy.power(rh, i + 2)
    return r


if __name__ == "__main__":


    print "####################################################################\n\nNote: WCA_FreeEnergy takes a few minutes to run\n\n####################################################################"

    gen_ref = (len(sys.argv) > 1 and sys.argv[1] == "gen_ref")

    T = 0.80
    WCA = WCA_FreeEnergy()
    WCA.rhoMin = 0.05
    WCA.rhoMax = 0.50
    WCA.rhoStep = 0.05
    
    WCA.MainLoop(T)


    nParams = 6
    # second and third virial coefficients of WCA potential at T=0.80
    B2, B3 = 2.25566, 3.15556

    infile = open("outputWCA%.3f.dat" % T)


    W0List = []
    toFitList = []
    nextLine = infile.readline()
    while nextLine:
        while nextLine.startswith("#"):
            nextLine = infile.readline()
        rho, meanW, W0, meanPE = map(string.atof, string.split(nextLine))
        minusKnownVir = meanW/T/rho - B2 - B3*rho
        toFitList.append([rho, minusKnownVir])
        W0List.append(W0)

        nextLine = infile.readline()
    
    # fit this to polynomial starting with rho**3
    startParams = []

    for pdx in xrange(nParams):
        startParams.append(0.)
    
    fitParams, chiSq = leastSquaresFit(PolyModel2, startParams, toFitList, stopping_limit=1.0e-6)
    print "chiSq",chiSq

    if gen_ref:
        outfile = open("free_energy_over_T_ref.dat","w")
    else:
        outfile = open("free_energy_over_T.dat","w")
        
    outfile.write("# rho total_F_over_T\n")

    if not gen_ref:
        sum_sq_diffs = 0
        infile = open("free_energy_over_T_ref.dat")
        infile.readline() # get the comment line out of the way

    for rdx in xrange(len(toFitList)):
        rho, valToFit = toFitList[rdx]
        fittedVal = PolyModel2(fitParams, rho)
        integral = B2*rho + B3*rho**2/2
        for pdx in xrange(nParams):
            integral += fitParams[pdx] * rho**(3+pdx)/(3.+pdx)

        totalFoverT = integral + W0List[rdx]/T
        outfile.write("%f %f\n" % (rho, totalFoverT))

        if not gen_ref:
            ref_rho, ref_FoverT = map(string.atof, string.split(infile.readline()))
            sum_sq_diffs += (totalFoverT - ref_FoverT)**2
    outfile.close()

    if not gen_ref:
        print "sum_sq_diffs",sum_sq_diffs
        if sum_sq_diffs < 1.e-3:
            sys.exit(0)
        else:
            sys.exit(1)
